from threading import local, Lock

from json_stream.dump import JSONStreamEncoder, _original_default

_lock = Lock()  # locked during state changes
_counter = 0  # number of patched threads
_thread = local()  # is *this* thread patched
_patched = Lock()  # locked while patch is active


class ThreadSafeJSONStreamEncoder(JSONStreamEncoder):
    def __enter__(self):
        global _counter
        with _lock:
            if _counter == 0:
                # patch if we are first
                _patched.acquire()
                super().__enter__()
            _thread.patched = True
            _counter += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        global _counter
        with _lock:
            _counter -= 1
            if _counter == 0:
                # unpatch if we are last
                super().__exit__(exc_type, exc_val, exc_tb)
                _patched.release()
            _thread.patched = False

    def default(self, obj):
        # if we end up being called by a thread that is _not_
        # patch (i.e. in a ThreadSafeJSONStreamEncoder
        # context), we must ensure that the patch is not active
        if not getattr(_thread, "patched", False):
            # block until any patching has been removed
            with _patched:
                # patch cannot be applied while in here
                assert not getattr(_thread, "patched", False)
                return _original_default
        return super().default(obj)



import json_stream
from json_stream.select_tokenizer import default_tokenizer


CONTENT_CHUNK_SIZE = 10 * 1024


def _to_iterable(response, chunk_size):
    return response.iter_bytes(chunk_size=chunk_size)


def load(response, persistent=False, tokenizer=default_tokenizer, chunk_size=CONTENT_CHUNK_SIZE):
    return json_stream.load(_to_iterable(response, chunk_size), persistent=persistent, tokenizer=tokenizer)


def visit(response, visitor, tokenizer=default_tokenizer, chunk_size=CONTENT_CHUNK_SIZE):
    return json_stream.visit(_to_iterable(response, chunk_size), visitor, tokenizer=tokenizer)



from json_stream.loader import load  # noqa: F401
from json_stream.visitor import visit  # noqa: F401
from json_stream.writer import streamable_list, streamable_dict  # noqa: F401
from json_stream.util import to_standard_types



import json
from io import BytesIO, StringIO
from itertools import zip_longest
from unittest import TestCase

from json_stream.select_tokenizer import default_tokenizer

from json_stream import load
from json_stream.base import TransientAccessException


class JSONLoadTestCase(TestCase):
    def _test_object(self, obj, persistent, binary=False, tokenizer=default_tokenizer):
        self.assertListEqual(list(self._to_data(obj, persistent, binary, tokenizer)), list(obj))
        self.assertListEqual(list(self._to_data(obj, persistent, binary, tokenizer).keys()), list(obj.keys()))
        self.assertListEqual(list(self._to_data(obj, persistent, binary, tokenizer).values()), list(obj.values()))
        self.assertListEqual(list(self._to_data(obj, persistent, binary, tokenizer).items()), list(obj.items()))

        for key in obj.keys():
            self.assertEqual(self._to_data(obj, persistent, binary, tokenizer).get(key), obj[key])

        self.assertFalse("foobar" in obj.keys())
        self.assertEqual(self._to_data(obj, persistent, binary, tokenizer).get("foobar"), None)
        self.assertEqual(self._to_data(obj, persistent, binary, tokenizer).get("foobar", "specified default"), "specified default")

        if persistent:
            self.assertEqual(len(self._to_data(obj, persistent, binary, tokenizer)), len(obj))
        for k, expected_k in zip_longest(self._to_data(obj, persistent, binary, tokenizer), obj):
            self.assertEqual(k, expected_k)

        if not persistent:
            data = self._to_data(obj, persistent, binary, tokenizer)
            iter(data)  # iterates first time
            with self.assertRaises(TransientAccessException):
                iter(data)  # can't get second iterator
            with self.assertRaises(TransientAccessException):
                data.keys()  # can't get keys
            with self.assertRaises(TransientAccessException):
                data.values()  # can't get keys
            with self.assertRaises(TransientAccessException):
                data.items()  # can't get keys

    def _test_list(self, obj, persistent, binary=False, tokenizer=default_tokenizer):
        self.assertListEqual(list(self._to_data(obj, persistent, binary, tokenizer)), list(obj))
        if persistent:
            self.assertEqual(len(self._to_data(obj, persistent, binary, tokenizer)), len(obj))
        for k, expected_k in zip_longest(self._to_data(obj, persistent, binary, tokenizer), obj):
            self.assertEqual(k, expected_k)

        if not persistent:
            data = self._to_data(obj, persistent, binary, tokenizer)
            iter(data)  # iterates first time
            with self.assertRaises(TransientAccessException):
                iter(data)  # can't get second iterator

    def _to_data(self, obj, persistent, binary, tokenizer):
        data = json.dumps(obj)
        if binary:
            stream = BytesIO(data.encode())
        else:
            stream = StringIO(data)
        return load(stream, persistent=persistent, tokenizer=tokenizer)



from io import StringIO
import json
from unittest import TestCase

import json_stream


class TestToStandardTypes(TestCase):
    JSON = '{"x": 1, "y": {}, "xxxx": [1,2, {"yyyy": 1}, "z", 1, []]}'

    def test_to_standard_types(self):
        js = json_stream.load(StringIO(self.JSON))
        converted = json_stream.to_standard_types(js)
        comparison = json.load(StringIO(self.JSON))
        self.assertEqual(converted, comparison)



import copy
import json
from io import StringIO

from json_stream import load
from json_stream.base import (
    TransientAccessException,
    PersistentStreamingJSONObject,
    TransientStreamingJSONList,
    TransientStreamingJSONObject,
    PersistentStreamingJSONList,
)
from json_stream.tests import JSONLoadTestCase


class TestLoader(JSONLoadTestCase):
    def test_load_empty_object(self):
        obj = {}
        self._test_object(obj, persistent=True)
        self._test_object(obj, persistent=False)

    def test_load_object(self):
        obj = {"a": 1, "b": None, "c": True}
        self._test_object(obj, persistent=True)
        self._test_object(obj, persistent=False)

    def test_load_object_binary(self):
        obj = {"a": 1, "b": None, "c": True}
        self._test_object(obj, persistent=True, binary=True)
        self._test_object(obj, persistent=False, binary=True)

    def test_load_object_get_persistent(self):
        json_str = '{"a": 1, "b": null, "c": true}'

        # Access in order
        data = load(StringIO(json_str), persistent=True)
        self.assertEqual(data['a'], 1)
        self.assertEqual(data['b'], None)
        self.assertEqual(data['c'], True)
        with self.assertRaises(KeyError):
            _ = data['d']

        # Access out of order
        data = load(StringIO(json_str), persistent=True)
        self.assertEqual(data['b'], None)
        self.assertEqual(data['a'], 1)
        self.assertEqual(data['c'], True)
        with self.assertRaises(KeyError):
            _ = data['d']

        # Access with key error first order
        data = load(StringIO(json_str), persistent=True)
        with self.assertRaises(KeyError):
            _ = data['d']
        self.assertEqual(data['a'], 1)
        self.assertEqual(data['b'], None)
        self.assertEqual(data['c'], True)

    def test_load_object_get_transient(self):
        json_str = '{"a": 1, "b": null, "c": true}'

        # Access in order
        data = load(StringIO(json_str), persistent=False)
        self.assertEqual(data['a'], 1)
        self.assertEqual(data['b'], None)
        self.assertEqual(data['c'], True)
        with self.assertRaises(TransientAccessException):
            _ = data['d']

        # Access out of order
        data = load(StringIO(json_str), persistent=False)
        self.assertEqual(data['b'], None)
        with self.assertRaises(TransientAccessException):
            _ = data['a']
        with self.assertRaises(TransientAccessException):
            _ = data['c']  # stream was exhausted in search for 'a'
        with self.assertRaises(TransientAccessException):
            _ = data['d']  # don't know if this was a key error or was in the past

        # Access with key error first order
        data = load(StringIO(json_str), persistent=False)
        with self.assertRaises(KeyError):
            _ = data['d']
        with self.assertRaises(TransientAccessException):
            _ = data['a']  # stream was exhausted in search for 'd'

    def test_load_empty_list(self):
        obj = []
        self._test_list(obj, persistent=True)
        self._test_list(obj, persistent=False)

    def test_load_list(self):
        obj = [1, True, ""]
        self._test_list(obj, persistent=True)
        self._test_list(obj, persistent=False)

    def test_load_list_get_persistent(self):
        json_str = '[1, true, ""]'

        # Access in order
        data = load(StringIO(json_str), persistent=True)
        self.assertEqual(data[0], 1)
        self.assertTrue(data[1])
        self.assertEqual(data[2], "")
        with self.assertRaises(IndexError):
            _ = data[3]

        # Access out of order
        data = load(StringIO(json_str), persistent=True)
        self.assertTrue(data[1])
        self.assertEqual(data[0], 1)
        self.assertEqual(data[2], "")
        with self.assertRaises(IndexError):
            _ = data[3]

    def test_load_list_get_transient(self):
        json_str = '[1, true, ""]'

        # Access in order
        data = load(StringIO(json_str), persistent=False)
        self.assertEqual(data[0], 1)
        self.assertTrue(data[1])
        self.assertEqual(data[2], "")
        with self.assertRaises(IndexError):
            _ = data[3]

        # Access out of order
        data = load(StringIO(json_str), persistent=False)
        self.assertTrue(data[1])
        with self.assertRaises(TransientAccessException):
            _ = data[0]
        self.assertEqual(data[2], "")
        with self.assertRaises(IndexError):
            _ = data[3]

    def test_load_nested_persistent(self):
        json_str = '{"count": 3, "results": ["a", "b", {}]}'
        data = load(StringIO(json_str), persistent=True)
        self.assertIsInstance(data, PersistentStreamingJSONObject)
        results = data['results']
        self.assertIsInstance(results, PersistentStreamingJSONList)
        self.assertEqual(results[0], 'a')
        self.assertEqual(results[1], 'b')
        self.assertIsInstance(results[2], PersistentStreamingJSONObject)
        self.assertEqual(len(results), 3)
        self.assertEqual(len(results[2]), 0)
        self.assertEqual(len(data), 2)
        self.assertEqual(data["count"], 3)

    def test_load_nested_transient(self):
        json_str = '{"count": 3, "results": ["a", "b", "c"]}'
        data = load(StringIO(json_str), persistent=False)
        self.assertIsInstance(data, TransientStreamingJSONObject)
        results = data['results']
        self.assertIsInstance(results, TransientStreamingJSONList)
        self.assertEqual(list(results), ['a', 'b', 'c'])

    def test_load_nested_transient_first_list_item_object(self):
        json_str = '[{"a": 4}, "b", "c"]'
        data = load(StringIO(json_str), persistent=False)
        self.assertIsInstance(data, TransientStreamingJSONList)
        items = iter(data)
        item = next(items)
        self.assertIsInstance(item, TransientStreamingJSONObject)
        self.assertDictEqual({"a": 4}, dict(item.items()))
        self.assertEqual(list(items), ['b', 'c'])

    def test_load_nested_transient_first_list_item_list(self):
        json_str = '[["a"], "b", "c"]'
        data = load(StringIO(json_str), persistent=False)
        self.assertIsInstance(data, TransientStreamingJSONList)
        items = iter(data)
        item = next(items)
        self.assertIsInstance(item, TransientStreamingJSONList)
        self.assertListEqual(["a"], list(item))
        self.assertEqual(list(items), ['b', 'c'])

    def test_not_copiable(self):
        json_str = '[["a"], "b", "c"]'
        with self.assertRaisesRegex(copy.Error, "^Copying json_steam objects leads to a bad time$"):
            copy.copy(load(StringIO(json_str)))
        with self.assertRaisesRegex(copy.Error, "^Copying json_steam objects leads to a bad time$"):
            copy.deepcopy(load(StringIO(json_str)))

    def test_transient_to_persistent(self):
        json_str = '{"results": [{"x": 1, "y": 3}, {"y": 4, "x": 2}]}'
        xs = iter((1, 2))
        ys = iter((3, 4))

        data = load(StringIO(json_str))  # data is a transient dict-like object
        self.assertIsInstance(data, TransientStreamingJSONObject)

        results = data['results']
        self.assertIsInstance(results, TransientStreamingJSONList)

        # iterate transient list, but produce persistent items
        for result in results.persistent():
            # result is a persistent dict-like object
            self.assertIsInstance(result, PersistentStreamingJSONObject)
            x = next(xs)
            y = next(ys)
            self.assertEqual(result['x'], x)
            self.assertEqual(result['y'], y)  # would error on second result without .persistent()
            self.assertEqual(result['x'], x)  # would error without .persistent()

    def test_persistent_to_transient(self):
        json_str = """{"a": 1, "x": ["long", "list", "I", "don't", "want", "in", "memory"], "b": 2}"""
        data = load(StringIO(json_str), persistent=True).transient()
        self.assertIsInstance(data, PersistentStreamingJSONObject)

        self.assertEqual(data["a"], 1)
        list_ = data["x"]
        self.assertIsInstance(list_, TransientStreamingJSONList)
        self.assertEqual(data["b"], 2)
        self.assertEqual(data["b"], 2)  # would error if data was transient
        with self.assertRaisesRegex(TransientAccessException, "Index 0 already passed in this stream"):
            _ = list_[0]  # cannot access transient list

    def test_primitives(self):
        primitives = (
            ("true", True),
            ("false", False),
            ("123", 123),
            ("12.3", 12.3),
            ('"string"', "string"),
        )
        for primitive, expected in primitives:
            result = load(StringIO(primitive))
            self.assertEqual(expected, result)

    def test_unterminated_list(self):
        with self.assertRaisesRegex(ValueError, "Unterminated list at end of file"):
            load(StringIO('[')).read_all()
        with self.assertRaisesRegex(ValueError, "Unterminated list at end of file"):
            load(StringIO('[1')).read_all()
        with self.assertRaisesRegex(ValueError, "Unterminated list at end of file"):
            load(StringIO('[1,')).read_all()

    def test_unterminated_object(self):
        with self.assertRaisesRegex(ValueError, "Unterminated object at end of file"):
            load(StringIO('{')).read_all()
        with self.assertRaisesRegex(ValueError, "Unterminated object at end of file"):
            load(StringIO('{"x"')).read_all()
        with self.assertRaisesRegex(ValueError, "Unterminated object at end of file"):
            load(StringIO('{"x":')).read_all()
        with self.assertRaisesRegex(ValueError, "Unterminated object at end of file"):
            load(StringIO('{"x": 1')).read_all()
        with self.assertRaisesRegex(ValueError, "Unterminated object at end of file"):
            load(StringIO('{"x": 1,')).read_all()

    def test_unicode(self):
        data = json.dumps('Ä')
        result = load(StringIO(data))
        self.assertEqual(result, 'Ä')

    def test_non_bmp_unicode(self):
        data = json.dumps('𝄞')
        result = load(StringIO(data))
        self.assertEqual(result, '𝄞')



from unittest import TestCase

from json_stream_rs_tokenizer import rust_tokenizer_or_raise

import json_stream
from json_stream import to_standard_types
from json_stream.tokenizer import tokenize


class TestBuffering(TestCase):
    def test_buffering(self):
        self._test_buffering(tokenizer=rust_tokenizer_or_raise())

    def test_buffering_python_tokenizer(self):
        self._test_buffering(tokenizer=tokenize)

    def _test_buffering(self, tokenizer):
        happenings = []

        def data_in_chunks(data, chunk_size=15):
            for i in range(0, len(data), chunk_size):
                part = data[i:i + chunk_size]
                happenings.append(('yield', part))
                yield part

        json_string = b'{"tasks":[{"id":1,"title":"task1"},{"id":2,"title":"task2"},{"id":3,"title":"task3"}]}'
        stream = json_stream.load(data_in_chunks(json_string), tokenizer=tokenizer)

        for task in stream["tasks"]:
            happenings.append(('item', to_standard_types(task)))

        self.assertListEqual([
            ('yield', b'{"tasks":[{"id"'),
            ('yield', b':1,"title":"tas'),
            ('yield', b'k1"},{"id":2,"t'),
            ('item', {'id': 1, 'title': 'task1'}),
            ('yield', b'itle":"task2"},'),
            ('item', {'id': 2, 'title': 'task2'}),
            ('yield', b'{"id":3,"title"'),
            ('yield', b':"task3"}]}'),
            ('item', {'id': 3, 'title': 'task3'})
        ], happenings)



import io
import json
from itertools import zip_longest
from unittest import TestCase
from unittest.mock import Mock

from json_stream.httpx import load, visit


class TestLoad(TestCase):
    maxDiff = None

    @staticmethod
    def grouper(iterable, n):
        """Collect data into fixed-length chunks or blocks"""
        args = [iter(iterable)] * n
        return zip_longest(*args, fillvalue="")

    def _create_mock_response(self):
        # httpx iter_bytes returns an iterable of bytes
        response = Mock()
        data = json.dumps({
            "a": "a" * io.DEFAULT_BUFFER_SIZE,
            "b": "b",
        })
        content = ("".join(chunk).encode() for chunk in self.grouper(data, 1024))
        response.iter_bytes.return_value = content
        return response

    def _assertDataOkay(self, data):
        # is streaming
        self.assertTrue(data.streaming)

        # check the data
        self.assertTupleEqual(
            (
                ("a", "a" * io.DEFAULT_BUFFER_SIZE),
                ("b", "b"),
            ),
            tuple(sorted(data.items())),
        )

        # all done?
        self.assertFalse(data.streaming)

    def test_load_persistent(self):
        response = self._create_mock_response()

        # load in persistent mode
        data = load(response, persistent=True)

        self._assertDataOkay(data)

    def test_load_transient(self):
        response = self._create_mock_response()

        # load in transient mode
        data = load(response, persistent=False)

        self._assertDataOkay(data)

    def test_visitor(self):
        response = self._create_mock_response()

        visited = []
        visit(response, lambda item, path: visited.append((item, path)))

        self.assertListEqual([
            ('a' * io.DEFAULT_BUFFER_SIZE, ('a',)),
            ('b', ('b',)),
        ], visited)



from collections import deque
from functools import wraps


class Streamable:
    def __init__(self, iterable):
        super().__init__()
        self._it = iter(iterable)
        self._cache = deque()

    def __iter__(self):
        return self

    def __next__(self):
        if self._cache:
            return self._cache.popleft()
        return next(self._it)

    def _peek(self):
        try:
            peek = next(self._it)
        except StopIteration:
            pass
        else:
            self._cache.append(peek)

    def __bool__(self):
        self._peek()
        return bool(self._cache)

    def __repr__(self):  # pragma: no cover
        return f'<{type(self).__name__} for {self._it}>'


class StreamableList(Streamable, list):
    """
        Class specifically designed to pass isinstance(o, list)
        and conform to the implementation of json.dump(o)
        for lists, except items are provided by passed in
        generator
    """


class StreamableDict(Streamable, dict):
    """
        Class specifically designed to pass isinstance(o, dict)
        and conform to the implementation of json.dump(o)
        for lists, except items are provided by passed in
        generator. Generator must produce pairs of key/value
    """
    def items(self):
        return self


def streamable_dict(fn):
    if not callable(fn):
        return StreamableDict(fn)

    @wraps(fn)
    def wrapper(*args, **kwargs):
        return StreamableDict(fn(*args, **kwargs))
    return wrapper


def streamable_list(fn):
    if not callable(fn):
        return StreamableList(fn)

    @wraps(fn)
    def wrapper(*args, **kwargs):
        return StreamableList(fn(*args, **kwargs))
    return wrapper


__all__ = ['streamable_dict', 'streamable_list']



import json
from io import StringIO
from unittest import TestCase

from json_stream.writer import streamable_dict, streamable_list


class TestWriter(TestCase):
    def test_writer_wrapper(self):
        def dict_content(n):
            yield "a  list", streamable_list(range(4))

        o = streamable_dict(dict_content(5))

        buffer = StringIO()
        json.dump(o, buffer)
        self.assertEqual('{"a  list": [0, 1, 2, 3]}', buffer.getvalue())

    def test_writer_decorator(self):

        @streamable_list
        def list_content(n):
            return range(n)

        @streamable_dict
        def dict_content(n):
            yield "a  list", list_content(n)

        o = dict_content(5)

        buffer = StringIO()
        json.dump(o, buffer)
        self.assertEqual('{"a  list": [0, 1, 2, 3, 4]}', buffer.getvalue())

    def test_writer_empty(self):

        @streamable_list
        def empty_list():
            for i in range(0):  # never yields
                yield i

        o = empty_list()

        buffer = StringIO()
        json.dump(o, buffer)
        self.assertEqual('[]', buffer.getvalue())



from warnings import warn

from json_stream.tokenizer import tokenize
from json_stream_rs_tokenizer import rust_tokenizer_or_raise, ExtensionException

try:
    default_tokenizer = rust_tokenizer_or_raise()
except ExtensionException as e:
    warn(str(e), category=ImportWarning)  # ImportWarnings are ignored by default
    default_tokenizer = tokenize

__all__ = ['default_tokenizer']



"""
Taken from the NAYA project

https://github.com/danielyule/naya

Copyright (c) 2019 Daniel Yule
"""
import re
from io import StringIO
from unittest import TestCase

from json_stream.tokenizer import tokenize, TokenType


class TestJsonTokenization(TestCase):
    def tokenize_sequence(self, string):
        return [token for token in tokenize(StringIO(string))]

    def assertNumberEquals(self, expected, actual):
        token_list = self.tokenize_sequence(actual)
        self.assertEqual(1, len(token_list))
        ttype, token = token_list[0]
        self.assertEqual(expected, token)
        self.assertEqual(ttype, TokenType.NUMBER)

    def assertOperatorEquals(self, expected, actual):

        token_list = self.tokenize_sequence(actual)
        ttype, token = token_list[0]
        self.assertEqual(expected, token)
        self.assertEqual(ttype, TokenType.OPERATOR)

    def assertStringEquals(self, *, expected, json_input):
        token_list = self.tokenize_sequence(json_input)
        self.assertEqual(1, len(token_list))
        ttype, token = token_list[0]
        self.assertEqual(expected, token)
        self.assertEqual(ttype, TokenType.STRING)

    def test_number_parsing(self):
        self.assertNumberEquals(0, "0")
        self.assertNumberEquals(0.5, "0.5")
        self.assertNumberEquals(0, "-0")
        self.assertNumberEquals(12, "12")
        self.assertNumberEquals(3.5, "3.5")
        self.assertNumberEquals(1.2e11, "12e10")
        self.assertNumberEquals(7.8e-14, "78E-15")
        self.assertNumberEquals(0, "0e10")
        self.assertNumberEquals(65.7, "65.7")
        self.assertNumberEquals(892.978, "892.978")
        self.assertNumberEquals(8.9e7, "8.9E7")
        self.assertRaises(ValueError, self.tokenize_sequence, "01")
        self.assertRaises(ValueError, self.tokenize_sequence, "1.")
        self.assertRaises(ValueError, self.tokenize_sequence, "-01")
        self.assertRaises(ValueError, self.tokenize_sequence, "2a")
        self.assertRaises(ValueError, self.tokenize_sequence, "-a")
        self.assertRaises(ValueError, self.tokenize_sequence, "3.b")
        self.assertRaises(ValueError, self.tokenize_sequence, "3.e10")
        self.assertRaises(ValueError, self.tokenize_sequence, "3.6ea")
        self.assertRaises(ValueError, self.tokenize_sequence, "67.8e+a")

    def test_operator_parsing(self):
        self.assertOperatorEquals("{", "{")
        self.assertOperatorEquals("}", "}")
        self.assertOperatorEquals("[", "[")
        self.assertOperatorEquals("]", "]")
        self.assertOperatorEquals(":", ":")
        self.assertOperatorEquals(",", ",")

    def test_string_parsing(self):
        self.assertStringEquals(expected="word", json_input=r'"word"')
        self.assertStringEquals(expected="with\tescape", json_input=r'"with\tescape"')
        self.assertStringEquals(expected="with\n a different escape", json_input=r'"with\n a different escape"')
        self.assertStringEquals(expected="using a \bbackspace", json_input=r'"using a \bbackspace"')
        self.assertStringEquals(expected="now we have \f a formfeed", json_input=r'"now we have \f a formfeed"')
        self.assertStringEquals(expected="\"a quote\"", json_input=r'"\"a quote\""')
        self.assertStringEquals(expected="", json_input=r'""')
        self.assertStringEquals(expected="/", json_input=r'"\/"')
        self.assertStringEquals(expected="this char at end: Ȃ", json_input=r'"this char at end: \u0202"')
        self.assertStringEquals(expected="this char in middle: Ȃ.", json_input=r'"this char in middle: \u0202."')
        self.assertStringEquals(expected="꽸", json_input=r'"\uaf78"')
        self.assertStringEquals(expected="訋", json_input=r'"\u8A0b"')
        self.assertStringEquals(expected="돧", json_input=r'"\uB3e7"')
        self.assertStringEquals(expected="ዯ", json_input=r'"\u12eF"')
        with self.assertRaisesRegex(ValueError, re.escape(r"Invalid unicode literal: \uay76 at index 6")):
            self.tokenize_sequence(r'"\uay76"')
        with self.assertRaisesRegex(ValueError, "Invalid string escape: h at index 2"):
            self.tokenize_sequence(r'"\h"')
        with self.assertRaisesRegex(ValueError, "Invalid string escape: 2 at index 2"):
            self.tokenize_sequence(r'"\2"')
        with self.assertRaisesRegex(ValueError, "Invalid string escape: ! at index 2"):
            self.tokenize_sequence(r'"\!"')
        with self.assertRaisesRegex(ValueError, "Unterminated unicode literal at end of file"):
            self.tokenize_sequence(r'"\u!"')

    def test_unterminated_strings(self):
        with self.assertRaisesRegex(ValueError, "Unterminated string at end of file"):
            self.tokenize_sequence('"unterminated')

    def test_sequence(self):
        result = self.tokenize_sequence("123 \"abc\":{}")
        self.assertEqual(result, [(2, 123), (1, 'abc'), (0, ':'), (0, '{'), (0, '}')])

        # Borrowed from http://en.wikipedia.org/wiki/JSON
        big_file = """{
          "firstName": "John",
          "lastName": "Smith",
          "isAlive": true,
          "isDead": false,
          "age": 25,
          "height_cm": 167.6,
          "address": {
            "streetAddress": "21 2nd Street",
            "city": "New York",
            "state": "NY",
            "postalCode": "10021-3100"
          },
          "phoneNumbers": [
            {
              "type": "home",
              "number": "212 555-1234"
            },
            {
              "type": "office",
              "number": "646 555-4567"
            }
          ],
          "children": [],
          "spouse": null
        }"""
        result = self.tokenize_sequence(big_file)
        expected = [(0, '{'), (1, 'firstName'), (0, ':'), (1, 'John'), (0, ','), (1, 'lastName'), (0, ':'),
                    (1, 'Smith'), (0, ','), (1, 'isAlive'), (0, ':'), (3, True), (0, ','), (1, 'isDead'), (0, ':'),
                    (3, False), (0, ','), (1, 'age'), (0, ':'),
                    (2, 25), (0, ','), (1, 'height_cm'), (0, ':'), (2, 167.6), (0, ','), (1, 'address'), (0, ':'),
                    (0, '{'), (1, 'streetAddress'), (0, ':'), (1, '21 2nd Street'), (0, ','), (1, 'city'), (0, ':'),
                    (1, 'New York'), (0, ','), (1, 'state'), (0, ':'), (1, 'NY'), (0, ','), (1, 'postalCode'),
                    (0, ':'), (1, '10021-3100'), (0, '}'), (0, ','), (1, 'phoneNumbers'), (0, ':'), (0, '['), (0, '{'),
                    (1, 'type'), (0, ':'), (1, 'home'), (0, ','), (1, 'number'), (0, ':'), (1, '212 555-1234'),
                    (0, '}'), (0, ','), (0, '{'), (1, 'type'), (0, ':'), (1, 'office'), (0, ','), (1, 'number'),
                    (0, ':'), (1, '646 555-4567'), (0, '}'), (0, ']'), (0, ','), (1, 'children'), (0, ':'), (0, '['),
                    (0, ']'), (0, ','), (1, 'spouse'), (0, ':'), (4, None), (0, '}')]
        self.assertListEqual(result, expected)
        big_file_no_space = '{"firstName":"John","lastName":"Smith","isAlive":true,"isDead":false,"age":25,"height_cm' \
                            '":167.6,"address":{"streetAddress":"21 2nd Street","city":"New York","state":"NY","posta' \
                            'lCode":"10021-3100"},"phoneNumbers":[{"type":"home","number":"212 555-1234"},{"type":"of' \
                            'fice","number":"646 555-4567"}],"children":[],"spouse":null}'
        result = self.tokenize_sequence(big_file_no_space)
        self.assertListEqual(result, expected)
        result = self.tokenize_sequence("854.6,123")
        self.assertEqual(result, [(2, 854.6), (0, ','), (2, 123)])
        self.assertRaises(ValueError, self.tokenize_sequence, "123\"text\"")
        self.assertRaises(ValueError, self.tokenize_sequence, "23.9e10true")
        self.assertRaises(ValueError, self.tokenize_sequence, "\"test\"56")

    def test_unicode_literal(self):
        result = list(tokenize(StringIO(r'"\u00c4"')))
        self.assertListEqual(result, [(1, 'Ä')])

    def test_unicode_literal_truncated(self):
        with self.assertRaisesRegex(ValueError, re.escape(r'Invalid unicode literal: \u00c" at index 6')):
            list(tokenize(StringIO(r'"\u00c"')))

    def test_unicode_literal_bad_hex(self):
        with self.assertRaisesRegex(ValueError, re.escape(r"Invalid unicode literal: \u00x4 at index 6")):
            list(tokenize(StringIO(r'"\u00x4"')))

    def test_unicode_surrogate_pair_literal(self):
        result = list(tokenize(StringIO(r'"\ud834\udd1e"')))
        self.assertListEqual(result, [(1, '𝄞')])

    def test_unicode_surrogate_pair_unpaired(self):
        with self.assertRaisesRegex(ValueError, "Unpaired UTF-16 surrogate at index 7"):
            list(tokenize(StringIO(r'"\ud834"')))
        with self.assertRaisesRegex(ValueError, "Unpaired UTF-16 surrogate at end of file"):
            list(tokenize(StringIO(r'"\ud834')))
        with self.assertRaisesRegex(ValueError, "Unpaired UTF-16 surrogate at index 8"):
            list(tokenize(StringIO(r'"\ud834\x')))
        with self.assertRaisesRegex(ValueError, "Unpaired UTF-16 surrogate at end of file"):
            list(tokenize(StringIO(r'"\ud834' + '\\')))

    def test_unicode_surrogate_pair_non_surrogate(self):
        with self.assertRaisesRegex(ValueError, "Second half of UTF-16 surrogate pair is not a surrogate! at index 12"):
            list(tokenize(StringIO(r'"\ud834\u00c4"')))

    def test_unicode_surrogate_pair_literal_truncated(self):
        with self.assertRaisesRegex(ValueError, re.escape(r'Invalid unicode literal: \u00c" at index 12')):
            list(tokenize(StringIO(r'"\ud834\u00c"')))

    def test_unicode_surrogate_pair_literal_bad_hex(self):
        with self.assertRaisesRegex(ValueError, re.escape(r"Invalid unicode literal: \u00x4 at index 12")):
            list(tokenize(StringIO(r'"\ud834\u00x4"')))

    def test_unicode_surrogate_pair_literal_invalid(self):
        message = re.escape(r"Error decoding UTF-16 surrogate pair \ud834\ud834 at index 12")
        with self.assertRaisesRegex(ValueError, message):
            list(tokenize(StringIO(r'"\ud834\ud834"')))

    def test_unicode_surrogate_pair_literal_unterminated(self):
        with self.assertRaisesRegex(ValueError, r"Unterminated unicode literal at end of file"):
            list(tokenize(StringIO(r'"\ud834\ud83')))



import json

from json_stream import to_standard_types
from json_stream.base import StreamingJSONBase

_original_default = json.JSONEncoder().default


class JSONStreamEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, StreamingJSONBase):
            return to_standard_types(obj)
        return _original_default(obj)

    def __enter__(self):
        json.JSONEncoder.default = self.default

    def __exit__(self, exc_type, exc_val, exc_tb):
        json.JSONEncoder.default = _original_default


default = JSONStreamEncoder().default



from collections import deque

from json_stream.base import StreamingJSONList, StreamingJSONObject


class Context:
    LIST = 1
    DICT = 2


def to_standard_types(x):
    in_stack, out_stack = deque(), deque()
    if isinstance(x, StreamingJSONList):
        in_stack.append((Context.LIST, iter(x)))
        output = []
    elif isinstance(x, StreamingJSONObject):
        in_stack.append((Context.DICT, iter(x.items())))
        output = {}
    else:
        return x
    out_stack.append(output)
    while len(in_stack) > 0:
        try:
            in_context, in_iter = in_stack[-1]
            in_elem = next(in_iter)
            if in_context == Context.LIST:
                if isinstance(in_elem, StreamingJSONList):
                    out_list = []
                    out_stack[-1].append(out_list)
                    out_stack.append(out_list)
                    in_stack.append((Context.LIST, iter(in_elem)))
                elif isinstance(in_elem, StreamingJSONObject):
                    out_dict = {}
                    out_stack[-1].append(out_dict)
                    out_stack.append(out_dict)
                    in_stack.append((Context.DICT, iter(in_elem.items())))
                else:
                    out_stack[-1].append(in_elem)
            elif in_context == Context.DICT:
                in_key, in_value = in_elem
                if isinstance(in_value, StreamingJSONList):
                    out_list = []
                    out_stack[-1][in_key] = out_list
                    out_stack.append(out_list)
                    in_stack.append((Context.LIST, iter(in_value)))
                elif isinstance(in_value, StreamingJSONObject):
                    out_dict = {}
                    out_stack[-1][in_key] = out_dict
                    out_stack.append(out_dict)
                    in_stack.append((Context.DICT, iter(in_value.items())))
                else:
                    out_stack[-1][in_key] = in_value
        except StopIteration:
            in_stack.pop()
            out_stack.pop()
    return output



from io import StringIO
from unittest import skipUnless

import json_stream_rs_tokenizer

from json_stream import load
from json_stream.tokenizer import tokenize

from json_stream.select_tokenizer import default_tokenizer

from json_stream.tests import JSONLoadTestCase


class TestPythonTokenizer(JSONLoadTestCase):
    def test_load_object(self):
        obj = {"a": 1, "b": None, "c": True}
        self._test_object(obj, persistent=False, tokenizer=tokenize)

    def test_load_object_binary(self):
        self.assertIs(default_tokenizer, json_stream_rs_tokenizer.RustTokenizer)
        obj = {"a": 1, "b": None, "c": True}
        self._test_object(obj, persistent=False, binary=True, tokenizer=tokenize)


@skipUnless(hasattr(json_stream_rs_tokenizer, 'RustTokenizer'), 'rust tokenizer not available')
class TestRSTokenizer(JSONLoadTestCase):
    def test_load_object(self):
        self.assertIs(default_tokenizer, json_stream_rs_tokenizer.RustTokenizer)
        obj = {"a": 1, "b": None, "c": True}
        self._test_object(obj, persistent=False)

    def test_load_object_binary(self):
        self.assertIs(default_tokenizer, json_stream_rs_tokenizer.RustTokenizer)
        obj = {"a": 1, "b": None, "c": True}
        self._test_object(obj, persistent=False, binary=True)

    def test_unterminated_string(self):
        with self.assertRaisesRegex(ValueError, "Unterminated string"):
            load(StringIO('["unterminated')).read_all()



import json
from io import StringIO
from unittest import TestCase

import json_stream
from json_stream.dump import default, JSONStreamEncoder


class TestDump(TestCase):
    JSON = '{"count": 3, "results": ["a", "b", "c"]}'

    def test_dump_default(self):
        data = json_stream.load(StringIO(self.JSON))
        output = json.dumps(data, default=default)
        self._assert_json_okay(output)

    def test_dump_cls(self):
        data = json_stream.load(StringIO(self.JSON))
        output = json.dumps(data, cls=JSONStreamEncoder)
        self._assert_json_okay(output)

    def test_dump_context(self):
        data = json_stream.load(StringIO(self.JSON))
        with JSONStreamEncoder():
            output = json.dumps(data)
        self._assert_json_okay(output)

    def test_dump_default_persistent(self):
        data = json_stream.load(StringIO(self.JSON), persistent=True)
        output = json.dumps(data, default=default)
        self._assert_json_okay(output)

    def test_dump_cls_persistent(self):
        data = json_stream.load(StringIO(self.JSON), persistent=True)
        output = json.dumps(data, cls=JSONStreamEncoder)
        self._assert_json_okay(output)

    def test_dump_context_persistent(self):
        data = json_stream.load(StringIO(self.JSON), persistent=True)
        with JSONStreamEncoder():
            output = json.dumps(data)
        self._assert_json_okay(output)

    def _assert_json_okay(self, value):
        self.assertEqual('{"count": 3, "results": ["a", "b", "c"]}', value)



import io
from unittest import TestCase
from json_stream.iterators import IterableStream


class TestIterableStream(TestCase):
    def test_read(self):
        # create some chunks of binary data
        data = (
            b"a" * io.DEFAULT_BUFFER_SIZE,
            b"b" * (io.DEFAULT_BUFFER_SIZE + 1),
            b"c" * (io.DEFAULT_BUFFER_SIZE - 1),
        )

        # stream it and check the result
        stream = IterableStream(data)
        self.assertEqual(stream.read(), b"".join(data))



from json_stream.base import StreamingJSONBase, TokenType
from json_stream.iterators import ensure_file
from json_stream.select_tokenizer import default_tokenizer


def load(fp_or_iterable, persistent=False, tokenizer=default_tokenizer):
    fp = ensure_file(fp_or_iterable)
    token_stream = tokenizer(fp)
    token_type, token = next(token_stream)
    if token_type == TokenType.OPERATOR:
        return StreamingJSONBase.factory(token, token_stream, persistent)
    return token
